/*
 * openE57las.h - header for simple interface for reading LAS files.
 *
 * Copyright (c) 2009 - 2010 Kevin Ackley (kackley@gwi.net)
 * Copyright (c) 2020 - 2022 Michele Adduci (adduci@tutanota.com)
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated by
 * a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
#ifndef LASREADER_INCLUDED
#define LASREADER_INCLUDED

#include <fstream>
#include <iostream>
#include <vector>
#ifndef E57FOUNDATIONIMPL_H_INCLUDED
#  include <openE57/impl/openE57Impl.h>
#endif

namespace e57
{
struct LASPublicHeaderBlock
{                                               //                   file format#:  1.0 1.1 1.2 1.3
  char     fileSignature[4];                    // required   4   4   4   4
  uint16_t fileSourceId;                        // required       2   2   2
  uint8_t  gpsTimeType;                         // required           1b  1b
  bool     waveformDataPacketsInternal;         // required               1b
  bool     waveformDataPacketsExternal;         // required               1b
  bool     returnNumbersSyntheticallyGenerated; // required               1b
  uint32_t projectIdGuidData1;                  // optional   4   4   4   4
  uint16_t projectIdGuidData2;                  // optional   2   2   2   2
  uint16_t projectIdGuidData3;                  // optional   2   2   2   2
  uint8_t  projectIdGuidData4[8];               // optional   8   8   8   8
  uint8_t  versionMajor;                        // required   1   1   1   1
  uint8_t  versionMinor;                        // required   1   1   1   1
  char     systemIdentifier[32];                // required  32  32  32  32
  char     generatingSoftware[32];              // required  32  32  32  32
  uint16_t fileCreationDayOfYear;               // *=reqd     2   2   2   2*
  uint16_t fileCreationYear;                    // *=reqd     2   2   2   2*
  uint16_t headerSize;                          // required   2   2   2   2
  uint32_t offsetToPointData;                   // required   4   4   4   4
  uint32_t numberOfVariableLengthRecords;       // required   4   4   4   4
  uint8_t  pointDataFormatId;                   // required   1   1   1   1
  uint16_t pointDataRecordLength;               // required   2   2   2   2
  uint32_t numberOfPointRecords;                // required   4   4   4   4
  uint32_t numberOfPointsByReturn[7];           // required  20  20  20  28
  double   xScaleFactor;                        // required   8   8   8   8
  double   yScaleFactor;                        // required   8   8   8   8
  double   zScaleFactor;                        // required   8   8   8   8
  double   xOffset;                             // required   8   8   8   8
  double   yOffset;                             // required   8   8   8   8
  double   zOffset;                             // required   8   8   8   8
  double   maxX;                                // required   8   8   8   8
  double   minX;                                // required   8   8   8   8
  double   maxY;                                // required   8   8   8   8
  double   minY;                                // required   8   8   8   8
  double   maxZ;                                // required   8   8   8   8
  double   minZ;                                // required   8   8   8   8
  uint64_t startOfWaveformDataPacketRecord;     // required               8

  /// Diagnostic functions:
  void dump(int indent = 0, std::ostream& os = std::cout);

  LASPublicHeaderBlock()
  {
    memset(this, 0, sizeof(*this));
  };
};

//              file format#:  1.0 1.1 1.0 1.1
struct LASPointDataRecord
{                                            //              point format#: pf0 pf0 pf1 pf1 pf2 pf3 pf4 pf5
  int32_t  x{0};                             // required, 32  32  32  32  32  32  32  32
  int32_t  y{0};                             // required, 32  32  32  32  32  32  32  32
  int32_t  z{0};                             // required, 32  32  32  32  32  32  32  32
  uint16_t intensity{0};                     // optional, 16  16  16  16  16  16  16  16
  uint8_t  returnNumber{0};                  // required,  3   3   3   3   3   3   3   3
  uint8_t  numberOfReturns{0};               // required,  3   3   3   3   3   3   3   3
  bool     scanDirectionFlag{false};         // required,  1   1   1   1   1   1   1   1
  bool     edgeOfFlightLine{false};          // required,  1   1   1   1   1   1   1   1
  uint8_t  classification{0};                // required,  8   5   8   5   5   5   5   5
  bool     synthetic{false};                 // required,      1       1   1   1   1   1
  bool     keyPoint{false};                  // required,      1       1   1   1   1   1
  bool     withheld{false};                  // required,      1       1   1   1   1   1
  int8_t   scanAngleRank{0};                 // required,  8   8   8   8   8   8   8   8
  uint8_t  fileMarker{0};                    // optional,  8       8
  uint16_t userBitField{0};                  // optional  16      16
  uint8_t  userData{0};                      // optional       8       8   8   8   8   8
  uint16_t pointSourceId{0};                 // required      16      16  16  16  16  16
  double   gpsTime{0.0};                     // required          64  64      64  64  64
  uint16_t red{0};                           // required                  16  16      16
  uint16_t green{0};                         // required                  16  16      16
  uint16_t blue{0};                          // required                  16  16      16
  uint8_t  wavePacketDescriptorIndex{0};     // required                           8   8
  uint64_t byteOffsetToWaveformData{0};      // required                          64  64
  uint32_t waveformPacketSizeInBytes{0};     // required                          32  32
  float    returnPointWaveformLocation{0.0}; // required                          32  32
  float    xT{0.0};                          // required                          32  32
  float    yT{0.0};                          // required                          32  32
  float    zT{0.0};                          // required                          32  32

  /// Diagnostic functions:
  void dump(int indent = 0, std::ostream& os = std::cout);
};

struct LASVariableRecordLengthHeader
{                                   //!!! not needed
  uint16_t reserved;                // 2 bytes  optional
  char     userId[16];              // 16 bytes required
  uint16_t recordId;                // 2 bytes  required
  uint16_t recordLengthAfterHeader; // 2 bytes  required
  char     description[32];         // 32 bytes optional

  /// Diagnostic functions:
  void dump(int indent = 0, std::ostream& os = std::cout);

  LASVariableRecordLengthHeader()
  {
    memset(this, 0, sizeof(*this));
  };
};

struct LASVariableLengthRecord
{
  ustring              userId;
  uint16_t             recordId;
  uint16_t             recordLengthAfterHeader;
  ustring              description;
  std::vector<uint8_t> wholeRecordData;

  /// Diagnostic functions:
  void dump(int indent = 0, std::ostream& os = std::cout);
};

struct LASGeoKeyHeader
{
  uint16_t keyDirectoryVersion;
  uint16_t keyRevision;
  uint16_t minorRevision;
  uint16_t numberOfKeys;

  /// Diagnostic functions:
  void dump(int indent = 0, std::ostream& os = std::cout);

  LASGeoKeyHeader()
  {
    memset(this, 0, sizeof(*this));
  };
};

struct LASGeoKeyEntry
{
  uint16_t keyId;
  uint16_t tiffTagLocation;
  uint16_t count;
  uint16_t valueOffset;

  /// Diagnostic functions:
  void dump(int indent = 0, std::ostream& os = std::cout);

  LASGeoKeyEntry()
  {
    memset(this, 0, sizeof(*this));
  };
};

class LASReader
{
public:
  LASReader(ustring fname);
  void operator~();
  void close(){/*???*/}; //???TODO implement

  void                 getHeader(LASPublicHeaderBlock& h);
  std::vector<uint8_t> getRawHeader();
  int                  readPoints(LASPointDataRecord* points, uint32_t start, uint32_t count);

  void rewindVLR();
  bool readNextVLR(LASVariableLengthRecord& vlrInfo);

  /// Diagnostic functions:
  void dump(int indent = 0, std::ostream& os = std::cout);

  /*!!!
      int readEVLR
      read raw
      key header, key contents
  */

protected: //================
  std::ifstream        fs_;
  LASPublicHeaderBlock hdr_{};
  unsigned             nextVLROffset_{0};
  unsigned             readVLRCount_{0};

  // TODO this doesn't work in MSVC6.0:
  // static const int POINT_DATA_RECORD_MAX_BYTES = 63;  // version dependency!, change if point record gets longer
#define POINT_DATA_RECORD_MAX_BYTES 63

  void readPoint(LASPointDataRecord& point);
};

} // namespace e57

#endif
